/* SimpleJetZ.java - applet for reading/writing of 16 byte MIFARE blocks */

package com.telenor.rd.nfc;

import javacard.framework.*;
import com.ibm.jz.JZSystem;

/**
 * A simple Java Card applet that provides ADPU commands for reading and writing of
 * 16 byte datablocks on the SmartMX MIFARE 1k emulation. 
 *
 * Only one 16 byte datablock can be read or written at a time. Valid datablocks are
 * in the range from 0 to 63 except:
 * 		block 0
 * 		trailer blocks
 * 
 * Read APDU:
 * 		CLA = 0x80, INS=0x01, P1/P2=0x00, LC=<length of CDATA>, CDATA=<data>
 * 			<data> : <block><mf-password>
 * Write APDU:
 * 		CLA = 0x80, INS=0x02, P1/P2=0x00, LC=<length of CDATA>, CDATA=<data>
 * 			<data> : <block><mf-password><16 byte data>
 * Application specific return values:
 * 		0x9000 (SW_NO_ERROR) 	 - no error, 16 bytes with data on read
 * 		0x6984 (SW_DATA_INVALID) - requested MIFARE block outside of range (0-63)
 * 		0x6A80 (SW_WRONG_DATA) 	 - error from JZSystem call, typically indicates wrong
 * 								   mf-password and/or access of a trailer block
 * 
 * 2006-02-26  Kjell M. Myksvoll, <kjell.myksvoll@telenor.com>
 */
public class SimpleJetZ extends javacard.framework.Applet {
	private static final byte JZ_CLA = (byte) 0x80;
	private static final byte JZ_INS_READ  = (byte) 0x01;
	private static final byte JZ_INS_WRITE = (byte) 0x02;

	public static void install(byte[] bArray, short bOffset, byte bLength) {
		(new SimpleJetZ()).register(bArray, (short)(bOffset + 1), bArray[bOffset]);
	}

	public void process(APDU apdu) {
		byte[] buf = apdu.getBuffer();

		// ignore select instruction
		if (buf[ISO7816.OFFSET_CLA] == 0x00 && buf[ISO7816.OFFSET_INS] == (byte) 0xA4)
			return;

		if (buf[ISO7816.OFFSET_CLA] != JZ_CLA)
			ISOException.throwIt(ISO7816.SW_CLA_NOT_SUPPORTED);
		if (buf[ISO7816.OFFSET_INS] != JZ_INS_READ && buf[ISO7816.OFFSET_INS] != JZ_INS_WRITE)
			ISOException.throwIt(ISO7816.SW_INS_NOT_SUPPORTED);

		short min_length = (buf[ISO7816.OFFSET_INS] == JZ_INS_READ)
				? (short) (1 + 8)
				: (short) (1 + 8 + 16);
		short lc = (short) buf[ISO7816.OFFSET_LC];

		if (lc < min_length || lc > (short) (buf.length - ISO7816.OFFSET_CDATA))
			ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);

		short block = (short) buf[ISO7816.OFFSET_CDATA];

		if (block > 63 || block < 0)
			ISOException.throwIt(ISO7816.SW_DATA_INVALID);

		if (buf[ISO7816.OFFSET_INS] == JZ_INS_READ) {
			byte[] resp = JCSystem.makeTransientByteArray((short) 24,
					JCSystem.CLEAR_ON_DESELECT);
			Util.arrayCopy(buf, (short) (ISO7816.OFFSET_CDATA + 1), resp, (short) 0, (short) 8);

			if (JZSystem.readWriteMifare(JZSystem.MIFARE_PASSWORD_READ, resp, (short) 0,
					block) != 0)
				ISOException.throwIt(ISO7816.SW_WRONG_DATA);

			apdu.setOutgoing();
			apdu.setOutgoingLength((short) 16);
			apdu.sendBytesLong(resp, (short) 8, (short) 16);
		} else {
			if (JZSystem.readWriteMifare(JZSystem.MIFARE_PASSWORD_WRITE, buf,
					(short) (ISO7816.OFFSET_CDATA + 1), block) != 0)
				ISOException.throwIt(ISO7816.SW_WRONG_DATA);

			ISOException.throwIt(ISO7816.SW_NO_ERROR);
		}
	}
}
