/**
 * Electronic Ticketing System (ETS)
 * Copyright (C) 2006, Magnus Egeberg
 * 
 * This file is part of ETS.
 * 
 * ETS is a program for using the mobile phone as a contactless electronic ticket.
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 */
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import com.nokia.nfc.sc.*;

/**
 * MifareWriter is responsible for communication with the java card application
 * 
 * @author Magnus Egeberg
 * @since 1.0
 */
public class MifareWriter implements Runnable{

	//Notifications to the user
	private final String CONFIRMATION_TITLE = "Confirmation";
	private final String BUSY_READING_TICKET = "Reading ticket...";
	private final String BUSY_SAVING_TICKET = "Saving ticket...";
	private final String BUSY_TICKET_SAVED = "Ticket saved and ready for use!";
	private final String BUSY_TICKET_READ = "Ticket read";
	
	//Used in reading and writing the ticket
	private byte[] buf = null;
	private byte[] newTotal = null;
	private byte[] readticket = null;
	private byte[] writeticket =  {(byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
								   (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
								   (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
								   (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00};
	
	//Counter to pause the thread
	private int count = 5;
	
	//Used in reading short values
	private int value;
	
	//Alert the user
	private ConfirmationAlert confAlert;
	
	//The handler of the secure chip
	private SecureChipHandler secureChip;
	
	//The MIDelt
	private TicketingSystemMIDlet ts;
	
	//The read / write task
	private boolean read = true;
	
	//The identification of the ticket
	private static final byte CMD_TICKET = (byte) 0xAA;
	
	//MIFARE password
	private static final byte[] MF_PASSWORD = {
	        (byte) 0x0B, (byte) 0x54, (byte) 0x57, (byte) 0x07,
	        (byte) 0x45, (byte) 0xFE, (byte) 0x3A, (byte) 0xE7
	};
	 
	 //select APDU for the correct JavaCard application
	 private static final byte[] SELECT = CommandAPDU.createSelectAPDU("SimpleJetZ.app").getApdu();

	/**
	 * Initiates the communication with the MIFARE memory
	 * 
	 * @param sc The handler of the communication with the secure chip
	 * @param ts The MIDlet
	 * @param data The new ticket
	 * @param read The task to be done, true when read
	 */ 
	public MifareWriter(SecureChipHandler sc, TicketingSystemMIDlet ts, byte[] data, boolean read){
		this.ts = ts;
		secureChip = sc;
		buf = data;
		this.read = read;
		confAlert = new ConfirmationAlert(CONFIRMATION_TITLE, ts);
		
    	new Thread(this).start();
	}
	
	/**
	 * Sets the communication mode of the secure chip to internal
	 * Identifies the operation, either read or write
	 */
	public void run() {
        ts.trace("run");
        //Waits for the cover to get connected
        try {
            while (secureChip.connected == false && count--> 0)
                Thread.currentThread().sleep(1000);
        } catch (Exception e) {
        	ts.trace("Exception RUN: " + e.toString());
            ts.showTrace();
        }
        
        if (secureChip.connected == true) {  
        	//Sets the secure chip communication mode to internal
        	if(secureChip.coverConnection.getSCCMode() != SCCMode.SC_INTERNAL){
        		secureChip.changeMode(SCCMode.SC_INTERNAL);
        	}
        	if (read == false){ 
        		//Write
        		update();
        	}
        	else{
        		//Read
        		ts.busy.setText(BUSY_READING_TICKET);
        		getTicket();
        	}
        }
        //The cover is disconnected
        else {
            ts.trace("cover disconnected, giving up!");
            ts.showTrace();
            sleep();
        }
      
    }
	
	/**
	 * Gets the existing credit on the secure chip
	 * and adds the credit from the new ticket
	 */
	private synchronized void update(){
         ts.trace("update");
         byte type = buf[1];
         int ticketLength = 0;
  
         try {
        	 //Select the SimpleJetZ java card application
             byte[] data = secureChip.coverConnection.exchangeAPDU(SELECT);
             
             String d = getHexString(data);
 			 ts.trace("APDU DATA: " + d);
             
             if (!(data.length == 2 && data[0] == (byte) 0x90 && data[1] == (byte) 0x00))
            	throw new APDUCommandException(data[0], data[1]);
             
             //Identify the ticket type
             if (type == CMD_TICKET) {
            	 ts.trace("type == CMD_TICKET");
                 ts.busy.setText(BUSY_SAVING_TICKET);
                 
                 //Get existing credit from secure chip
                 writeticket = readFromChip(61);
                 byte[] ticket = new byte[16];
                 ticketLength = buf.length - 2;
                 
                 //Check that the read ticket has correct length
                 if(ticketLength <= 16){
                	 System.arraycopy(buf, 2, ticket, 0, ticketLength);
                	 
                	 //Generate new total
                     byte[] currentValue = {writeticket[5], writeticket[6]};
                     byte[] newValue = {ticket[3], ticket[4]};
                     int currentCredit = getShort(currentValue);
                     int addedCredit = getShort(newValue);
                     int newCredit = currentCredit + addedCredit;
                     
                     //puts the new credit into newTotal[], byte 0 and 1.
                     writeShort(newCredit);
                     System.arraycopy(newTotal, 0, ticket, 5, 2);
                     
                     //Write the new ticket to secure chip
                     writeToChip(61, ticket);	 
                 }
                 else{
                	 ts.trace("Ticketlenght error: " + ticketLength +" bytes");
                	 ts.showTrace();
                 }  
             } 
        }
        catch (IOException e) {
        ts.trace("IOException UP: " + e.toString());
        ts.showTrace();
        }
        catch (APDUCommandException e) {
           ts.trace("APDUCommandException UP: " + e.toString());
           ts.showTrace();
        }
        catch (Exception e) {
           ts.trace("Exception UP: " + e.toString());
           ts.showTrace();
         }
         finally {
            try {
            	//Restore communication mode
            	ts.trace("Restoring communication mode");
            	secureChip.changeMode(SCCMode.SC_RESTORE);
             } catch (Exception e) {
                ts.trace("IOException UP: " + e.toString());
                ts.showTrace();
                //Put the thread to sleep
                sleep();
             }
         }
         if (ts.handle_error == false) {
        	 ts.trace("handle_error == false");

            if (type == CMD_TICKET) {
            	 confAlert.setText(BUSY_TICKET_SAVED);
            	 //Enter the main menu
            	 ts.display.setCurrent(confAlert, ts.mainForm);
         		ts.manager.playSequence(ts.vibeSequence);
         		ts.manager.setBacklight(1000 * 5);
         		//Put the thread to sleep
         		sleep();
             }
         }   
     }

	/**
	 * Writes to the MIFARE memory
	 * 
	 * @param pos The block number to write
	 * @param ticket The ticket to write
	 * 
	 * @throws IOException
	 * @throws APDUCommandException
	 * @throws Exception
	 */
     private void writeToChip(int pos, byte[] ticket)
         throws IOException, APDUCommandException, Exception {
    	 
    	 ts.trace("writeToChip");
         ts.trace("write: pos=" + pos + " len=" + ticket.length);

         //Control that the data is 16 bytes long, as this is what SimpleJetZ.app accepts
         if ((ticket.length % 16) != 0)
             throw new Exception("data not multiple of 16");

         //Control that the data is not addressed to a sector trailor;
         if(pos % 16 == 3 || pos == 0)
        	 throw new Exception("illegal destination block: " + pos);
         
         //Creaye the data to put into the APDU
         byte[] block = new byte[1 + 8 + 16];
         //Copy the MIFARE pw into the data
         System.arraycopy(MF_PASSWORD, 0, block, 1, 8);
         //Copy the address into the data
         block[0] = (byte) pos;
         //Copy the ticket into the data
         System.arraycopy(ticket, 0, block, 9, 16);
         
         //Generate the write APDU 
         CommandAPDU c = new CommandAPDU((byte) 0x80, (byte) 0x02, (byte) 0x00,
                                             (byte) 0x00, -1, block, 0, block.length);
            
         //The actual write APDU sent to the secure chip
         byte[] data = secureChip.coverConnection.exchangeAPDU(c.getApdu());
                  
         String tmp = getHexString(data);
       	 ts.trace("w.cmd.respons " + tmp +"\n");
         
       	 //Control the response
         if (!(data.length == 2 && data[0] == (byte) 0x90 && data[1] == (byte) 0x00)){	
                 throw new APDUCommandException(data[0], data[1]);
         }
     }
     
     /**
      * Reads the existing ticket
      */
     public void getTicket(){
    	 ts.trace("getTicket");
    	 //Select the SimpleJetZ java card application
         try {     	
        	 byte[] data = secureChip.coverConnection.exchangeAPDU(SELECT);
             
        	 if (!(data.length == 2 && data[0] == (byte) 0x90 && data[1] == (byte) 0x00))
        		 throw new APDUCommandException(data[0], data[1]);
        	
        	 //Reads the ticket from the secure chip
            readticket = readFromChip(61);
        
            ts.busy.setText(BUSY_TICKET_READ);
        }catch (IOException e) {
        	ts.trace("IOException UP: " + e.toString());
        	ts.showTrace();
        }catch (APDUCommandException e) {
            ts.trace("APDUCommandException UP: " + e.toString());
            ts.showTrace();
        }catch (Exception e) {
            ts.trace("Exception UP: " + e.toString());
            ts.showTrace();
         }
         finally {
            try {
            	 ts.trace("Restoring communication mode");
            	 //Restore communication mode
            	 secureChip.changeMode(SCCMode.SC_RESTORE);
             } catch (Exception e) {
                ts.trace("IOException UP: " + e.toString());
                ts.showTrace();
                //Put the thread to sleep
                sleep();
             }
         }
         if (ts.handle_error == false) {
        	 ts.trace("handle_error == false");
        	 //Present the ticet to the user
        	 new TicketForm(ts, readticket);
        	 //Put the thread to sleep
        	 sleep();
         }
         
     }
     
     /**
      * Reads the ticket from the MIFARE memory
      * 
      * @param pos The block # to read
      * @return The ticket that is read
      * @throws IOException
      * @throws APDUCommandException
      */
     public byte[] readFromChip(int pos)throws IOException, APDUCommandException{
    	 ts.trace("readFromChip");
    	 ts.trace("read: pos=" + pos);
    	 
    	 //Generate data to put in the request APDU
         byte[] block = new byte[1 + 8];
         //Copy MIFARE pw to data
         System.arraycopy(MF_PASSWORD, 0, block, 1, 8);
         //Copy address to data
         block[0] = (byte) pos;
         //Generate read APDU
         CommandAPDU c = new CommandAPDU((byte) 0x80, (byte) 0x01, (byte) 0x00,
                                             (byte) 0x00, -1, block, 0, block.length);
         
         //The actual read APDU sent to the secure chip
       	 byte[] data = secureChip.coverConnection.exchangeAPDU(c.getApdu());
           
       	 int length = data.length;
       	 ts.trace("resonse length: " +length);
       	 
       	 //Control the response
       	 if (length > 2){
       		 String tmp = getHexString(data);
           	 ts.trace("Stored ticket: " + tmp +"\n");
       	 }
         if (!( data.length == 18 && data[length-2] == (byte) 0x90 && data[length-1] == (byte) 0x00)){	
        	 throw new APDUCommandException(data[0], data[1]);
         }
         return data;
     }
    
     /**
 	 * Converts a byte array to string representation
 	 * 
 	 * @param data The byte array to 
 	 * @return the string representation of the byte array
 	 */
	private static String getHexString(byte[] data) {
		StringBuffer sb = new StringBuffer();
	        
	    for (int i = 0; i < data.length; i++) {
	    	String bs = Integer.toHexString(data[i] & 0xFF);
	        if (bs.length() == 1) sb.append(0);
	        	sb.append(bs);
	            sb.append((i + 1) % 4 == 0 ? '\n' : ' ');
	        }   
	        return sb.toString();
	    }
		
	/**
	 * Reads a short and return it as an int
	 * 
	 * @return The int value of the bytes
	 * @see ByteArrayInputStream
	 * @see DataInputStream
	 * @throws Exception when failing to read the short
	 */
	private int getShort(byte[] data){
		ts.trace("getShort");
		ByteArrayInputStream bais = new ByteArrayInputStream(data);
		DataInputStream in = new DataInputStream(bais);
		
		try {
			value = (int) in.readUnsignedShort();
	        in.close();
		    bais.close();
		}catch (IOException e) {
	        	ts.trace("IOException e: " + e.toString());
	            ts.showTrace();
	    }
		return value;
	}
		
	/**
	 * Writes an int to a short
	 * 
	 * @param theShort The int value to write
	 * @see ByteArrayOutputStream
	 * @see DataOutputStream
	 * @throws Exception
	 */
	public void writeShort(int theShort) throws Exception{
		ts.trace("writeShort");
		 ByteArrayOutputStream baos  = new ByteArrayOutputStream();
		 DataOutputStream out = new DataOutputStream(baos);
	     if((theShort < 0)||(theShort > 65535)){
	       	ts.trace("Short number out of range");
	        ts.showTrace();
	     }
	     try {
	    	 out.writeShort(theShort);
	     } catch (IOException e) {
	       	ts.trace("IOException e: " + e.toString());
	        ts.showTrace();
	     }
	     if(baos == null){
	    	 ts.trace("Not possible to get array from ByteArrayInputStream");
	         ts.showTrace();
	     }
	     //The new total ammount	
	     newTotal = baos.toByteArray();
	        
	     try {
	    	out.close();
		    baos.close();
	     } catch (Exception e) {
	        ts.trace("MifareWriter writeShort close error! " + e.toString());
	        ts.showTrace();
	     }
	}
	
	/**
	 * Puts the current thread to sleep
	 */	
	private void sleep(){
		boolean sleep = true;
		while(sleep){
			try{
				Thread.currentThread().sleep(10 * 1000);
			} catch(InterruptedException ie){		
				//Do nothing
			}
		}
	}
		
		
}
