/**
 * Electronic Ticketing System (ETS)
 * Copyright (C) 2006, Magnus Egeberg
 * 
 * This file is part of ETS.
 * 
 * ETS is a program for using the mobile phone as a contactless electronic ticket.
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 */
import java.io.IOException;
import javax.microedition.lcdui.*;

import com.nokia.nfc.sc.*;

/**
 * SecureChipHandler is responsible for communication with the
 * Nokia NFC payment and ticketing cover.
 * 
 * @author Magnus Egeberg
 * @since 1.0
 */
public class SecureChipHandler implements SecureChipListener, CommandListener {
       
	//Alert regarding PIN
	private final String ALERT_PIN_TITLE = "PIN";
	private final String ALERT_PIN_RECEIVED = "PIN received.";
	private final String ALERT_PIN_QUERY = "Please enter PIN.";
	private final String ALERT_PIN_ERROR_PIN_MISMATCH = "PIN mismatch";
	
	//Busy alert
	private final String BUSY_CORRECT_PIN = "Correct PIN.";
	
	//Commands
	private final String CMD_OK = "Ok";
	private final String CMD_BACK = "Back";
    
	//PIN change responses
    private final String MODE_CHANGE = "Mode change :";
	private final String MODE_CHANGE_SUCCESS = "Successful";
	private final String ERROR_PIN_MISMATCH = "PIN mismatch";
	private final String ERROR_PIN_TIMEOUT = "PIN timeout is pending";
	private final String ERROR_PIN_RESET_TIMEOUT = "PIN is reset, wait 24 hours";
	private final String ERROR_SC_BUSY = "The SC is busy with another session";
	 
	//Byte array containing the ticket that shall be written to the secure chip
	private byte[] buf = null;
	
	//The pending mifare task
	private boolean mifareWrite = false;
	
	//The PIN query
	private Alert pinAlert;
	private TextBox pinEntry;  
	private Command okCmd;
	private Command backCmd;
	
	//The connection to the cover
	public SecureChipConnection coverConnection;
	public boolean connected = false;
	public boolean pinOk = false;
	private String pin = null;
    	
	//The MIDlet
    private TicketingSystemMIDlet ts;
    
	//An empty ticket passes to the MifareWriter when read
	private byte[] empty =  {(byte) 0x00,(byte) 0x00,(byte) 0x00,(byte) 0x00,
							 (byte) 0x00,(byte) 0x00,(byte) 0x00,(byte) 0x00,
							 (byte) 0x00,(byte) 0x00,(byte) 0x00, (byte) 0x00,
							 (byte) 0x00,(byte) 0x00,(byte) 0x00, (byte) 0x00};
   
    /**
     * The SecureChipHandler creates the PIN alert and connects to the cover
     * 
     * @param ts The MIDlet.
     */
     public SecureChipHandler(TicketingSystemMIDlet ts) {
        	
    	 this.ts = ts;
         pinAlert = new Alert(ALERT_PIN_TITLE, ALERT_PIN_RECEIVED, null, AlertType.INFO);
         pinAlert.setTimeout(Alert.FOREVER);
         connectToCover();
     }
       
     /**
      * Handles events sent from the cover
      * 
      * @param e the SecureChipEvent sent from the cover
      * @see SecureChipEvent
      */
     public void handleEvent(SecureChipEvent e){
    	 ts.trace("handleEvent");
    	 if(e.getType() == SecureChipEvent.TYPE_CONNECTION_STATUS){
    			if(e.getStatus() == SecureChipEvent.STATUS_CONNECTED){
    				connected = true;
    				ts.trace("cover connected");
    			}else{
    				connected = false;
    				ts.trace("cover disconnected");
    			}
    	 }	
     }    
    	
     /**
      * changes the communication mode of the cover
      * 
      * @param mode The new communication mode
      * @see SecureChipConnection
      */  		
     public void changeMode(SCCMode mode){
    	 ts.trace("changeMode");
    	 boolean force = true;
    	 int timeout = 0;
    	 int responseCode;
    	 String response;
    	
    	 try{
    		 responseCode = coverConnection.setSCCMode(mode, force, pin, timeout);
         } catch(Exception e){
        	 ts.trace(e.toString());
             return;
         }
        		
         response = mapResult(responseCode);
         if(response == MODE_CHANGE_SUCCESS){
        	 StringItem changeModeItem = new StringItem(MODE_CHANGE, MODE_CHANGE_SUCCESS);
        	 ts.errorForm.append(changeModeItem);
         }else{
        	 ts.trace(response);
             ts.showTrace();
         }
     }
    
     /**
      * Query the user for PIN to change the communication mode
      * 
      * @param text The query
      */
     private void askForPin(String text){
    	ts.trace("askForPin");
    	
		//Creates the textbox
    	pinEntry = new TextBox(text, "", 4, TextField.PASSWORD | TextField.NUMERIC);
		
    	//Creates and adds the commands
    	okCmd = new Command(CMD_OK, Command.SCREEN, 1);
		backCmd = new Command(CMD_BACK, Command.BACK, 1);
		pinEntry.addCommand(okCmd);
		pinEntry.addCommand(backCmd);
		
		//Sets the commandsListener and makes the textbox visible
		pinEntry.setCommandListener(this);
		ts.display.setCurrent(pinEntry);
		ts.manager.setBacklight(1000 * 5);
	}
    
    /**
     * Updates the error form with data
     * 
     * @param parameter The parameter to trace
     * @param value The value of the parameter
     */
    public void updateForm(String parameter, String value){
    	ts.trace("updateForm");
    	StringItem info = new StringItem(parameter, value);
    	ts.errorForm.append(info);
   	}

    /**
     * Maps the given code to a error string.
     * 
     * @param The code to be mapped
     */
    private String mapResult(int responseCode) {
    	switch (responseCode) {
    		case SecureChipConnection.SUCCESS:
    			return MODE_CHANGE_SUCCESS;
    		case SecureChipConnection.ERROR_PIN_MISMATCH:
                return ERROR_PIN_MISMATCH;
            case SecureChipConnection.ERROR_PIN_TIMEOUT:
                return ERROR_PIN_TIMEOUT;
            case SecureChipConnection.ERROR_PIN_RESET_TIMEOUT:
                return ERROR_PIN_RESET_TIMEOUT;
            case SecureChipConnection.ERROR_SC_BUSY:
                return ERROR_SC_BUSY;
            default:
                return "Unknown error: " + responseCode;
    	}
    }
    	 
    /**
     * Performs actions based on the commands
     * 
     * @param c The command
     * @param d The displayable
     */	 
    public void commandAction(Command c, Displayable d){
    	ts.trace("sc: commandAction");
    	int response;
    	
    	//If the user selects OK
    	if (c == okCmd){
    		pinAlert.setTimeout(2000);
    		ts.display.setCurrent(pinAlert);
    		pin = pinEntry.getString();
    		
    		//Verify the PIN
    		try {
    			response = coverConnection.verifyPIN(pin);
    			if(response == SecureChipConnection.ERROR_PIN_MISMATCH){
    				askForPin(ALERT_PIN_ERROR_PIN_MISMATCH + ALERT_PIN_QUERY);
    				return;
    			}
    		} catch(IOException e){
    			ts.trace("PIN command");
    			ts.showTrace();
    		}
    		//The PIN is accepted
    		pinOk = true;
    		ts.busy.setText(BUSY_CORRECT_PIN);
    		ts.display.setCurrent(ts.busy);
    		
    		//If the task was to write	
    		if(mifareWrite == true){
    			mifareWrite = false;	
    			new MifareWriter(this, ts, buf, false);
    		}
    		//If the task was to read
    		else{
    			new MifareWriter(this, ts, empty, true);
    		}
    	}
    	//If the user selects back
    	else if(c == backCmd){
    		ts.display.setCurrent(pinAlert);
    	}
    }
    /**
     *Creats a MifareWriter.
     *If PIN is not ok, the user will be queried for the PIN
     * 
     * @param data The ticket
     */
    public void writeMifare(byte[] data){
    	ts.trace("writeMifare");
    	int count = 5;
    	//To give the SecureChipHandler time to connect to cover
    	while(connected == false && count-->0){
			try{
				Thread.currentThread().sleep(2000);
			} catch(Exception e){
				//Shall not be interrupted
			}
		}	
    	mifareWrite = true;
    	//Sets the buffer to 0 before copying the new ticket
    	if (buf != null){
    		resetTicket();
    	}
    	buf = data;
    	
    	//Checks if PIN is needed to change communication mode
    	if(pinEnabled() == true){
    		//Checks if PIN is ok
    		if(pinOk == true){
    			mifareWrite = false;
    			new MifareWriter(this, ts, data, false);
    		}
    		else{
    			askForPin(ALERT_PIN_QUERY);
    		}
    	}//PIN is not needed
    	else{
    		mifareWrite = false;
    		new MifareWriter(this, ts, data, false);
    	}
    }
    
    /**
     * Creates a MifareWriter
     * If PIN is not ok, the user will be queried for the PIN
     */
    public void readMifare(){
    	ts.trace("readMifare");
    	//Checks if PIN is needed to change communication mode
    	if(pinEnabled() == true){
    		//Checks if PIN is ok
    		if(pinOk == true){
    			new MifareWriter(this, ts, empty, true);
    		}else{
    			askForPin(ALERT_PIN_QUERY);
    		}
    	}//PIN is not needed
    	else{
    		new MifareWriter(this, ts, empty, true);
    	}
   	}
    
    /**
     * Controls of PIN is enabled
     * 
     * @return tmp True if PIN is enabled
     */
    private boolean pinEnabled(){
    	ts.trace("pinEnabled");
    	boolean tmp = coverConnection.isPinEnabled(coverConnection.getSCCMode());
    	return (tmp);
    }
    
    /**
     * Connects to the cover
     */
    public void connectToCover(){
    	ts.trace("connectToCover");
    	coverConnection = SecureChipConnection.openConnection(ts.getAppProperty("nfc-emulator-url"), this);
    }
    
    /**
     * Sets the buffer that the ticket is copied into all (byte) 0x00
     */
    private void resetTicket(){
    	ts.trace("resetTicket");
    	int length = buf.length;
    	for (int i=0 ; i<length ; i++){
    		buf[i]= (byte) 0x00;
    	}
    }
}
    
   