/**
 * Electronic Ticketing System (ETS)
 * Copyright (C) 2006, Magnus Egeberg
 * 
 * This file is part of ETS.
 * 
 * ETS is a program for using the mobile phone as a contactless electronic ticket.
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 */

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;

import javax.microedition.lcdui.*;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Displayable;

/**
 * TicketForm is responsible for presenting ticket data to the user.
 * 
 * @author Magnus Egeberg
 * @since 1.0
 */
public class TicketForm extends Form implements CommandListener{

	//Text presented to the user
	private final String CMD_BACK = "Back";
	private final String CMD_EXIT = "Exit";
	private final String TITLE_BALANCE = "Balance";
	private final String TITLE_PREVIOUS_CHARGE = "Previous charge";
	private final String INFO_BALANCE = "Current balance";
	private final String INFO_DATE = "Date";
	private final String INFO_AMOUNT = "Amount";
	private final String ENDING_KR = ",- kr.\n";
	
	//Commands
	private Command tBack;
	private Command tExit;
	
	//Streams to read bytes
	private ByteArrayInputStream bais;
	private DataInputStream in;
	
	//The MIDlet
	private TicketingSystemMIDlet ts;
	
	/**
	 * TicketForm creastes the form and adds the commands.
	 * 
	 * @param ts The MIDlet
	 * @param data The byte array that will be presented to the user
	 * @see ByteArrayInputStream
	 * @see DataInputStream
	 */
	public TicketForm(TicketingSystemMIDlet ts, byte[] data){
		super("TicketForm");
		this.ts = ts;
		tBack = new Command (CMD_BACK, Command.BACK, 2);
		tExit = new Command (CMD_EXIT, Command.EXIT, 1);
		addCommand(tBack);
		addCommand(tExit);
		setCommandListener(this);
		
		bais = new ByteArrayInputStream(data);
		in = new DataInputStream(bais);
		
		int day = 0;
		int month = 0;
		int year = 2000;
		int prevCharge = 0;
		int credit = 0;
		try{
			day = readByte();
			month = readByte();
			year += readByte();
			prevCharge = readShort();
			credit = readShort();
		}catch(Exception e){
			ts.trace(e.toString());
			ts.showTrace();
		}
		
		if(ts.info == 0){
			StringItem balance = new StringItem(INFO_BALANCE, credit+ENDING_KR);
			append(balance);
			setTitle(TITLE_BALANCE);
		} else if(ts.info == 1){
			StringItem date = new StringItem(INFO_DATE, day +"."+ month +"."+ year+".\n");
			StringItem prevAmmount = new StringItem(INFO_AMOUNT, prevCharge+ENDING_KR);
			append(date);
			append(prevAmmount);
			setTitle(TITLE_PREVIOUS_CHARGE);
		}
		ts.display.setCurrent(this);
   	 	ts.manager.setBacklight(1000 * 5);
	}
	
	/**
	 * Perfoms actions based on user commands
	 */
	public void commandAction(Command c, Displayable d){
		
		if (c == tExit)
			ts.destroyApp(false);
		if (c == tBack)
			ts.display.setCurrent(ts.mainForm);
	}
	
	/**
	 * Reads a short and return it as an int
	 * 
	 * @return The int value of the bytes
	 * @throws Exception when failing to read the short
	 */
	 public int readShort() throws Exception{
	        try {
	            return (int) in.readShort();
	        } catch (IOException e) {
	           throw new Exception("ByteArray read short error! \n"+e.getMessage());
	        }
	    }
	/**
	 * Reads a byte and returns it as an int
	 * 
	 * @return the int value of the byte
	 * @throws Exception when failing to read the byte
	 */
	 public int readByte() throws Exception{
		
	        try {
	            return (int) in.readUnsignedByte();
	        } catch (IOException e) {
	           throw new Exception("ByteArray read byte error! \n"+e.getMessage());
	        }
	    }
}
